﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Collections.Generic;

namespace _3dlifegameof
{
    // rules can be read at
    // http://www.math.com/students/wonders/life/life.html
    public class GenerationGenerator
    {

        private List<Point> CurrentGeneration;
        private int width;
        private int height;
        private bool init = false;
        public int Count;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="_CurrentGeneration">list of lifeforms</param>
        /// <param name="_width"></param>
        /// <param name="_height"></param>
        public GenerationGenerator(List<LifeForm> _CurrentGeneration, int _width, int _height) {
            CurrentGeneration = ConvertToPointList(_CurrentGeneration);
            width = _width;
            height = _height;
            Count = 0;
            init = true;
        }

        /// <summary>
        /// Compute one generation
        /// </summary>
        /// <returns>List of points</returns>
        public List<Point> OneGeneration(int _width, int _height) {
            if (!init) {
                throw new Exception("GenerationGenerator object has not be initialised");
            }
            width = _width;
            height = _height;
            Count++;
            List<Point> NewGeneration = new List<Point>();
            for (int y = 0; y < height; y++) {
                for (int x = 0; x < width; x++) {

                    // law 1: A dead cell with exactly three live neighbors becomes a live cell (birth). 
                    if (!GetPoint(x, y)) {
                        if (Law1(x, y)) {
                            NewGeneration.Add(new Point(x, y));
                        }
                    }
                    else {// law 2: A live cell with two or three live neighbors stays alive (survival). 
                        if (Law2(x, y)) {
                            NewGeneration.Add(new Point(x, y));
                        }
                        else {
                            // law 3: In all other cases, a cell dies or remains dead (overcrowding or loneliness). 
                            // no code, just do not add anything for the next generation
                        }
                    }
                }
            }
            CurrentGeneration = NewGeneration;
            return NewGeneration;
        }

        /// <summary>
        /// Return true is P is a new point in supplied point list
        /// </summary>
        /// <param name="NewGeneration"></param>
        /// <param name="P"></param>
        /// <returns></returns>
        private bool IsNew(List<Point> NewGeneration,Point P) {
            foreach (Point NGp in NewGeneration) {
                if (NGp.X == P.X && NGp.Y ==P.Y) return false;
            }
            return true;
        }

        /// <summary>
        /// Returns a Point list after obtaining all points from within a Lifeform list
        /// </summary>
        /// <param name="_CurrentGeneration"></param>
        /// <returns></returns>
        private List<Point> ConvertToPointList(List<LifeForm> _CurrentGeneration) {
            List<Point> Points = new List<Point>();
            foreach (LifeForm LF in _CurrentGeneration) Points.Add(LF.location);
            return Points;
        }

        // law 2: A live cell with two or three live neighbors stays alive (survival). 
        private bool Law2(int x, int y) {
            int N = Neighbours(x, y);
            if ((N == 3)||(N==2)) return true;
            return false;
        }

        // law 1: A dead cell with exactly three live neighbors becomes a live cell (birth). 
        private bool Law1(int x, int y) {
            if (Neighbours(x, y) == 3) return true;
            return false;
        }

        /// <summary>
        /// Returns the number of neighbouring cells
        /// area edges wrap around
        /// </summary>
        /// <param name="px"></param>
        /// <param name="py"></param>
        /// <returns></returns>
        private int Neighbours(int px, int py) {
            int c = 0;
            int Ox;
            int Oy;
            for (int y = -1; y < 2; y++) {
                for (int x = -1; x < 2; x++) {
                    if (!(x == 0 && y == 0)) {
                        Ox = px + x;
                        Oy = py + y;
                        if (Ox >= width) Ox = 0;
                        if (Ox < 0) Ox = width-1;
                        if (Oy >= height) Oy = 0;
                        if (Oy < 0) Oy = height-1;
                        if (GetPoint(Ox, Oy)) {
                            c++;
                        }
                    }
                }
            }
            return c;
        }

        /// <summary>
        /// Return true if x,y exists in the Current generation
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        private bool GetPoint(int x, int y) {
            foreach (Point P in CurrentGeneration) {
                if (P.X == x && P.Y == y) return true;
            }
            return false;
        }
    }


}
