﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Windows.Threading;

using Kit3D.Windows.Controls;
using Kit3D.Windows.Media;
using Kit3D.Windows.Media.Media3D;

namespace _3dlifegameof
{
    // 3D Life Game Of 2009 Rob Davis
    // My first foray into 3D with Silverlight
    // 3D support is supplied by Kit3D v0.3 http://www.codeplex.com/Kit3D
    // Reuses the JogWheel control from Positech2000 http://www.pay4foss.org/jumpstation/Positech2000/ 
    // Uses the original code for the blocks demo from Kit3D
    public partial class Page : UserControl
    {
        List<LifeForm> LifeForms = new List<LifeForm>(); // All the current LifeForms

        private Random random = new Random();
        private GenerationGenerator LifeSimulator;  // Class that handles the Life rules

        private ModelVisual3D modelVisualGrid;
        private Viewport3D viewport;
        private PerspectiveCamera Camera;
        private double AngleOffset = -0.0;
        private Point3D gridCenterPosition;
        private double distanceToCenter;
        private double orbitangle = 0;
        private double orbitX = 0;
        private double orbitZ = 0;
        private int GameWidth = 10;
        private int RegenCount = 0;
        private const int RegenMax = 3;
        private const int DelayMin = 10;
        private const int DelayMax = 3610;
        DispatcherTimer dt = new DispatcherTimer();
        private int StatExtinctGen = 0;
        private int StatTotalGen = 0;
        private int StatTLFB = 0;
        private int StatTLFD = 0;
        private int MilliSeconds=500;

        public Page(String GridSize) {
            InitializeComponent();
            GameWidth = Convert.ToInt32(GridSize);
            SpeedJogWheel.Update_Event += new JogWheel.Update_EventHandler(SpeedJogWheel_Update_Event);
            SpeedJogWheel.Restrict = true;
            SpeedJogWheel.ResetAngle(180);
            this.Loaded += new RoutedEventHandler(Prepare3DProjection);
            dt.Interval = new TimeSpan(0, 0, 0,0,500);
            dt.Tick += new EventHandler(dt_Tick);
            dt.Start();
        }

        /// <summary>
        /// Event to handle updates from the JogWheel control
        /// </summary>
        void SpeedJogWheel_Update_Event(object sender, MouseEventArgs e) {
            dt.Stop();
            Double SpeedSelected = SpeedJogWheel.Range;
            if (SpeedSelected >= 0.87) MilliSeconds = 3000;
            else if (SpeedSelected >= 0.75) MilliSeconds = 2000;
            else if (SpeedSelected >= 0.63) MilliSeconds = 1500;
            else if (SpeedSelected >= 0.5) MilliSeconds = 1000;
            else if (SpeedSelected >= 0.37) MilliSeconds = 400;
            else if (SpeedSelected >= 0.25) MilliSeconds = 120;
            else if (SpeedSelected >= 0.12) MilliSeconds = 60;
            else if (SpeedSelected >= 0.0) MilliSeconds = 30;
            
            dt.Interval = new TimeSpan(0, 0, 0, 0, MilliSeconds);
            dt.Start();
            UpdateStatsDisplay();
        }

        /// <summary>
        /// Core action area of the code
        /// gets a new Life generation and displays it
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void dt_Tick(object sender, EventArgs e) {
            dt = (DispatcherTimer)sender;
            List<Point> NewGeneration = new List<Point>();
            try {
                NewGeneration = LifeSimulator.OneGeneration(GameWidth, GameWidth);
            }
            catch  {
                LifeSimulator = new GenerationGenerator(LifeForms, GameWidth, GameWidth);
                NewGeneration = LifeSimulator.OneGeneration(GameWidth, GameWidth);
            }
            StatTotalGen++;
            bool ModifiedGen = false;
            foreach (Point P in NewGeneration) {
                if (!InLifeFormList(P)) ModifiedGen = true;
            }
            if (!ModifiedGen) RegenCount++;
            else RegenCount = 0;
            // foreach old gen point if not in new gen remove
            List<Point> RemovablePoints = new List<Point>();
            foreach (LifeForm LF in LifeForms) {
                if (!InPointList(NewGeneration, LF.location)) RemovablePoints.Add(LF.location);
            }
            while (RemovablePoints.Count > 0) {
                RemoveLifeForm(RemovablePoints[0].X, RemovablePoints[0].Y);
                RemovablePoints.RemoveAt(0);
            }
            // foreach new gen point not in old gen add
            foreach (Point P in NewGeneration) {
                if (!InLifeFormList(P)) {
                    DrawCube(P.X,P.Y, Colors.Blue);
                }
            }
            GenerationCount.Text = LifeSimulator.Count.ToString();
            if (RegenCount >= RegenMax) {
                RegenCount = 0;
                RandomlySeed();
                StatExtinctGen++;
            }
            UpdateStatsDisplay();
        }

        /// <summary>
        /// Update the Stats
        /// </summary>
        private void UpdateStatsDisplay() {
            TBStatExtinctGen.Text = StatExtinctGen.ToString();
            TBStatTLFB.Text = StatTLFB.ToString();
            TBStatTLFD.Text = StatTLFD.ToString();
            TBStatGenPerSec.Text = String.Format("{0:0.0}", (1000.0 / (Double)MilliSeconds));
            TBStatGameArea.Text = String.Format("{0} x {0}",GameWidth);
            TBStatTotalGen.Text = StatTotalGen.ToString();
        }

        /// <summary>
        /// Randomly re 'seed' the area with new lifeforms
        /// </summary>
        private void RandomlySeed() {
            LifeSimulator = null; // force the creation of a new simulator
            int RandomSeeds = random.Next(GameWidth);
            RandomSeeds += (int)(GameWidth / 2); // ensure there is at least a few seeds
            Point P;
            while (RandomSeeds > 0) {
                P = new Point((double)random.Next(GameWidth),(double)random.Next(GameWidth));
                if (!InLifeFormList(P)) DrawCube(P.X,P.Y,Colors.Red) ;
                RandomSeeds--;
            }
        }

        /// <summary>
        /// Creates a new Cube and draws it to the viewport
        /// applying transforms translation and colour
        /// </summary>
        private void DrawCube(double x, double y, Color color) {
            int sourceX = (int)x;
            int sourceY = (int)y;
            x *= 2;
            y *= 2;
            x = (-x)-2;
            Model3D Cube = GetCubeModel(color);
            Transform3DGroup transforms = new Transform3DGroup();
            TranslateTransform3D move = new TranslateTransform3D(new Vector3D(x, 0, y));
            transforms.Children.Add(move);
            Cube.Transform = transforms;
            ModelVisual3D visual = new ModelVisual3D();
            visual.Content = Cube;
            viewport.Children.Add(visual);
            LifeForms.Add(new LifeForm(visual, sourceX, sourceY));
            StatTLFB++;
        }

        /// <summary>
        /// Define a Cube, as 6 walls
        /// </summary>
        /// <returns></returns>
        private MeshGeometry3D CreateCubeMesh() {
            MeshGeometry3D mesh = new MeshGeometry3D();
            mesh.Positions = new Point3DCollection
            {
                new Point3D(0,0,0),
                new Point3D(1,0,0),
                new Point3D(1,0,1),
                new Point3D(0,0,0),
                new Point3D(1,0,1),
                new Point3D(0,0,1),
                new Point3D(0,1,0),
                new Point3D(1,1,1),
                new Point3D(1,1,0),
                new Point3D(0,1,0),
                new Point3D(0,1,1),
                new Point3D(1,1,1),
                new Point3D(0,0,0),
                new Point3D(0,1,0),
                new Point3D(1,1,0),
                new Point3D(0,0,0),
                new Point3D(1,1,0),
                new Point3D(1,0,0),
                new Point3D(1,0,0),
                new Point3D(1,1,0),
                new Point3D(1,1,1),
                new Point3D(1,0,0),
                new Point3D(1,1,1),
                new Point3D(1,0,1),
                new Point3D(1,0,1),
                new Point3D(1,1,1),
                new Point3D(0,1,1),
                new Point3D(1,0,1),
                new Point3D(0,1,1),
                new Point3D(0,0,1),
                new Point3D(0,0,1),
                new Point3D(0,1,1),
                new Point3D(0,1,0),
                new Point3D(0,0,1),
                new Point3D(0,1,0),
                new Point3D(0,0,0)
            };

            mesh.TriangleIndices = new Int32Collection
            {
                0,1,2,3,4,5,
                6,7,8,9,10,11,
                12,13,14,15,16,17,
                18,19,20,21,22,23,
                24,25,26,27,28,29,
                30,31,32,33,34,35
            };

            return mesh;
        }

        /// <summary>
        /// Define a flat plane as two triangles
        /// </summary>
        /// <returns></returns>
        private MeshGeometry3D CreateGridLineMesh() {
            MeshGeometry3D mesh = new MeshGeometry3D();
            mesh.Positions = new Point3DCollection
            {
                new Point3D(0,0,0),
                new Point3D(0.2,0,0),
                new Point3D(0.2,GameWidth*4,0),
                new Point3D(0,0,0),
                new Point3D(0,GameWidth*4,0),
                new Point3D(0.2,GameWidth*4,0),
            };

            mesh.TriangleIndices = new Int32Collection
            {
                0,1,2,5,4,3            };

            return mesh;
        }

        private Model3DGroup GridGroup;

        /// <summary>
        /// Add objects in a grid pattern
        /// </summary>
        /// <param name="width"></param>
        /// <param name="depth"></param>
        /// <param name="step">distance between lines</param>
        /// <returns></returns>
        private ModelVisual3D LoadGrid(int width, int depth, double step) {
            GridGroup = new Model3DGroup();
            double currentX = 0;
            double currentY = 0;
            double currentZ = 0;

            for (int z = 0; z <= depth; z++) {
                Model3D GridLine = GetGridLineModel(Colors.Green);
                GridLine.Transform = GetGridLineTransforms(new Vector3D(currentX, currentY, currentZ),false);
                GridGroup.Children.Add(GridLine);
                currentZ += step;
            }
            currentX = 0;
            currentY = 0;
            currentZ = 0;
            for (int z = 0; z <= width; z++) {
                Model3D GridLine = GetGridLineModel(Colors.Red);
                GridLine.Transform = GetGridLineTransforms(new Vector3D(currentX, currentY, currentZ), true);
                GridGroup.Children.Add(GridLine);
                currentX -= step;
            }
            ModelVisual3D visual = new ModelVisual3D();
            visual.Content = GridGroup;
            return visual;
        }


        private Transform3DGroup GetGridLineTransforms(Vector3D finalTranslateVector, bool axis) {
            Transform3DGroup transforms = new Transform3DGroup();
            TranslateTransform3D shrink = new TranslateTransform3D(-.5, 0, -.5);
            ScaleTransform3D scale = new ScaleTransform3D(new Vector3D(.5, .5, .5));
            TranslateTransform3D move = new TranslateTransform3D(finalTranslateVector);
            RotateTransform3D rotate = new RotateTransform3D(new AxisAngleRotation3D(new Vector3D(0, 0, 1), 90));
            transforms.Children.Add(rotate);
            rotate = new RotateTransform3D(new AxisAngleRotation3D(new Vector3D(-1, 0, 0), 90));
            if (axis) {
                transforms.Children.Add(rotate);
                rotate = new RotateTransform3D(new AxisAngleRotation3D(new Vector3D(0, 1, 0), 90));
            }
            transforms.Children.Add(rotate);
            transforms.Children.Add(shrink);
            transforms.Children.Add(scale);
            transforms.Children.Add(move);
            return transforms;

        }

        /// <summary>
        /// Turns the cube definition into a geometrical object
        /// </summary>
        private Model3D GetCubeModel(Color colour) {
            GeometryModel3D geometry = new GeometryModel3D();
            geometry.Geometry = CreateCubeMesh();
            geometry.SeamSmoothing = 1;
            Material colorMaterial = new DiffuseMaterial(new Kit3DBrush(new SolidColorBrush(colour)));
            geometry.Material = colorMaterial;
            return geometry;
        }

        /// <summary>
        /// Pass through
        /// </summary>
        /// <returns></returns>
        private Model3D GetGridLineModel() {
            return (GetGridLineModel(Colors.White));
        }

        /// <summary>
        /// Turns the object definition into a geometrical object
        /// </summary>
        /// <param name="c"></param>
        /// <returns></returns>
        private Model3D GetGridLineModel(Color c) {
            GeometryModel3D geometry = new GeometryModel3D();
            geometry.Geometry = CreateGridLineMesh();
            geometry.SeamSmoothing = 0.8;
            Material colorMaterial = new DiffuseMaterial(new Kit3DBrush(new SolidColorBrush(c)));
            geometry.Material = colorMaterial;
            return geometry;
        }

        /// <summary>
        /// Set up the camera and other 3D requirements
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void Prepare3DProjection(object sender, RoutedEventArgs e) {
            viewport = new Viewport3D();
            viewport.HorizontalAlignment = HorizontalAlignment.Stretch;
            viewport.VerticalAlignment = VerticalAlignment.Stretch;
            modelVisualGrid = LoadGrid(GameWidth, GameWidth, 2.0);
            Setup3DView();
            DrawCube(4, 4, Colors.Red); // this pattern is known as "R-pentomino"
            DrawCube(4, 5, Colors.Red); // and is a good starting point
            DrawCube(5, 4, Colors.Red); // as it seems so random
            DrawCube(4, 6, Colors.Red); // branching all over the place
            DrawCube(3, 5, Colors.Red);
            viewport.ShowModelBoundingBoxes = true;
            this.LayoutRoot.Children.Add(viewport);
            Kit3D.Windows.Media.CompositionTarget.Rendering += new EventHandler(CompositionTarget_Rendering);
        }

        /// <summary>
        /// if GameWidth changes then call this
        /// </summary>
        void Setup3DView() {
            Camera = new PerspectiveCamera();
            distanceToCenter = (GameWidth * 2) + 30;
            gridCenterPosition = new Point3D(-(GameWidth), 0, GameWidth);
            Camera.UpDirection = new Vector3D(0, 1, 0);
            Camera.FieldOfView = 45;
            Camera.Position = new Point3D(0, GameWidth * 2, 0);
            Camera.Position = Orbit(Camera.Position);
            viewport.Camera = Camera;
            Vector3D LookDirection = new Vector3D();
            //LookDirection.X = ToRadians(orbitangle) + AngleOffset ;
            //LookDirection.Y = -1.9;
            //LookDirection.Z = -0.9;
            LookDirection = new Vector3D(gridCenterPosition.X - Camera.Position.X,
            gridCenterPosition.Y - Camera.Position.Y,
            gridCenterPosition.Z - Camera.Position.Z);
            Camera.LookDirection = LookDirection;
            if (viewport.Children.Contains(modelVisualGrid)) {
                viewport.Children.Remove(modelVisualGrid);
                modelVisualGrid = LoadGrid(GameWidth, GameWidth, 2.0);
                viewport.Children.Add(modelVisualGrid);
            }
            else {
                modelVisualGrid = LoadGrid(GameWidth, GameWidth, 2.0);
            }
        }

        /// <summary>
        /// Event triggered by Kit3D for rotating 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void CompositionTarget_Rendering(object sender, EventArgs e) {
            Camera.Position = Orbit(Camera.Position);
            Vector3D LookDirection = new Vector3D();
            LookDirection.X = ToRadians(orbitangle) + AngleOffset;
            LookDirection.Y = Camera.LookDirection.Y;
            LookDirection.Z = Camera.LookDirection.Z;
            //Camera.LookDirection = LookDirection;
            LookDirection = new Vector3D(gridCenterPosition.X - Camera.Position.X,
            gridCenterPosition.Y - Camera.Position.Y,
            gridCenterPosition.Z - Camera.Position.Z);
            Camera.LookDirection = LookDirection;
            viewport.Camera = Camera;
        }

        /// <summary>
        /// Return a point from a circle
        /// </summary>
        /// <param name="P3d"></param>
        /// <returns></returns>
        private Point3D Orbit(Point3D P3d) {
            double rangle;
            rangle = ToRadians(orbitangle);
            P3d.X = -15+Math.Cos(rangle) * distanceToCenter + orbitX;
            P3d.Z = 10+Math.Sin(rangle) * distanceToCenter + orbitZ;
            orbitangle = orbitangle + 0.7;
            if (orbitangle > 360) {
                orbitangle = orbitangle - 360;
            }
            return P3d;
        }


        /// <summary>
        /// From degrees to radians
        /// </summary>
        /// <param name="angle">angle</param>
        /// <returns>angle in radians</returns>
        private static double ToRadians(double angle) {
            return (angle * Math.PI / 180);
        }

        /// <summary>
        /// from radians to degrees
        /// </summary>
        /// <param name="Radians">radians</param>
        /// <returns>angle in degrees</returns>
        private static double FromRadians(double Radians) {
            return (Radians * 180 / Math.PI);
        }

        /// <summary>
        /// Not currently used, allowed updates to static camera
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Button_Click(object sender, RoutedEventArgs e) {
            gridCenterPosition.X = Int32.Parse(txtLAX.Text);
            gridCenterPosition.Y = Int32.Parse(txtLAY.Text);
            gridCenterPosition.Z = Int32.Parse(txtLAZ.Text);
            distanceToCenter = Int32.Parse(txtD2C.Text);
            orbitX = Int32.Parse(txtORBX.Text);
            orbitZ = Int32.Parse(txtORBY.Text);
        }

        /// <summary>
        /// Allow user to force new lifeforms
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AddButton_Click(object sender, RoutedEventArgs e) {
            RandomlySeed();
            UpdateStatsDisplay();
        }

        /// <summary>
        /// Remove Lifeforms from the visual display and the internal storage
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        private void RemoveLifeForm(double x, double y) {
            List<LifeForm> ToDelete = new List<LifeForm>();
            foreach (LifeForm LF in LifeForms) {
                if (LF.location.X == x && LF.location.Y == y) ToDelete.Add(LF);
            }
            while (ToDelete.Count > 0) {
                viewport.Children.Remove(ToDelete[0].modelVisual);
                LifeForms.Remove(ToDelete[0]);
                ToDelete.Remove(ToDelete[0]);
            }
            StatTLFD++;
        }

        /// <summary>
        /// Return true if point is in Point list
        /// </summary>
        /// <param name="PL"></param>
        /// <param name="P"></param>
        /// <returns></returns>
        private bool InPointList(List<Point>PL,Point P) {
            foreach (Point p in PL) {
                if (p.X == P.X && p.Y == P.Y) return true;
            }
            return false;
        }

        /// <summary>
        /// Return true if Point in Lifeform list
        /// </summary>
        /// <param name="P"></param>
        /// <returns></returns>
        private bool InLifeFormList(Point P) {
            foreach (LifeForm LF in LifeForms) {
                if (LF.location.X == P.X && LF.location.Y == P.Y) return true;
            }
            return false;
        }


        private void SmallerButton_Click(object sender, RoutedEventArgs e) {
            if (GameWidth > 10) GameWidth -= 5;
            Cull(5);
            Setup3DView();
        }

        /// <summary>
        /// remove lifeforms outside the grid (due to manual shrinkage)
        /// </summary>
        private void Cull(int size) {
            List<LifeForm> ToDelete = new List<LifeForm>();
            foreach (LifeForm LF in LifeForms) {
                if (LF.location.X < size) ToDelete.Add(LF);
                if (LF.location.Y < size) ToDelete.Add(LF);
                if (LF.location.X > GameWidth) ToDelete.Add(LF);
                if (LF.location.Y < GameWidth) ToDelete.Add(LF);
            }
            while (ToDelete.Count > 0) {
                viewport.Children.Remove(ToDelete[0].modelVisual);
                LifeForms.Remove(ToDelete[0]);
                ToDelete.Remove(ToDelete[0]);
                StatTLFD++;
            }
        }

        private void BiggerButton_Click(object sender, RoutedEventArgs e) {
            if (GameWidth < 100) GameWidth += 5;
            Setup3DView();
        }

        private void ToggleGridButton_Click(object sender, RoutedEventArgs e) {
            if(viewport.Children.Contains(modelVisualGrid)) viewport.Children.Remove(modelVisualGrid); 
            else viewport.Children.Add(modelVisualGrid);
        }


    }


}
